# 
# Copyright (c) 2012, Kay-Uwe (Kiwi) Lorenz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of "Kay-Uwe (Kiwi) Lorenz" nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
'''module to access variables defined in calling frames

Inspired by Perl6's ``$*foo`` variables, which are "dynamically overridable
global variables" (`Synopsis 2 on Twigils`_), I have created this module.

.. _Synopsis 2 on Twigils: http://perlcabal.org/syn/S02.html#Twigils

Basically this enables you to access local variables from calling frames::

    >>> def a(): x = 1 ; return b()
    >>> def b(): return ctx('x')
    >>> a()
    1

If you rely on this feature and you set a variable *x*, which should be 
accessed in a called function, you quickly hit the problem, that your *x*
might be used in another called function locally and you do not get the 
expected value in your called function, but the one of the other *x*::

    >>> def a(): x = 1 ; return b()
    >>> def b(): x = 2 ; return c()
    >>> def c(): # expects x from a 
    ...     return ctx('x')
    >>> a()
    2

So it is good to brand context variables with some characters, which cannot
be used in usual python syntax for variables::

    >>> def a(): ctx('*x', 1) ; return b()
    >>> def b(): x = 2 ; return c()
    >>> def c(): # expects *x from a
    ...    return ctx('*x')
    >>> a()
    1

If you try to access a variable, which cannot be found, there will be raised
a NameError::

    >>> ctx('***does not exist***')
    Traceback (most recent call last):
    ...
    NameError: name '***does not exist***' not defined in calling frames

You can use :func:`ctxget` for specifying a default in such a case:

    >>> ctxget('***does not exist***', 'default')
    'default'

There are two convenience functions for normal import::

    >>> import ctxvar
    >>> ctxvar.set('*foo', 'bar')
    >>> ctxvar.get('*foo', 'default')
    'bar'

If you access a context var more than once within a function, it may be 
helpful to copy it to local frame, while getting the value::

    >>> def a(): ctx('*x', 1) ; return b()
    >>> def b(): x = 2 ; return c()
    >>> def c(): # expects *x from a
    ...    first = ctx('*x', Here)
    ...    return ctx('*x') # this will access the local copy
    >>> a()
    1

    >>> def a(): FOO = "Hello" ; return b()
    >>> def b(): BAR = "World" ; return c()
    >>> def c(): return ctxstr("${FOO} ${BAR}")
    >>> a()
    'Hello World'

This will go back the frames only once.

'''
__all__        = [ 'ctx', 'ctxget', 'ctxstr', 'Here' ]
__copyright__  = "Copyright 2012, Kay-Uwe (Kiwi) Lorenz"
__author__     = "Kay-Uwe (Kiwi) Lorenz"
__email__      = "kiwi@franka.dyndns.org"
__license__    = "BSD-3"
__status__     = "Production"
__version__    = '1.0'

import sys

if hasattr(sys, '_getframe'):
    getframe = sys._getframe
else:
    def _getframe():
        try:
            raise Exception
        except:
            return sys.exc_info()[2].tb_frame.f_back
    getframe = _getframe

class Undef:
    def __repr__(self): return 'Undef'

class Here:
    def __repr__(self): return 'Here'

def find_ctx_var(name, count=1, frame=Undef, caching=None):
    """return value of variable named *name* in calling frames

    :param name:
        name of the variable to be found
    :type name:
        String

    :param count:
        return value of *count*th found variable, default 1
    :type count:
        Integer

    :param frame:
        frame where to start searching, default is Undef, which means
        it will be automatically determined.
    :type frame:
        frame_

    :param caching:
        flag if the variable should be cached in each frame on the way to the
        variable.

    .. _frame: http://docs.python.org/reference/datamodel.html#frame-objects

    """
    if frame is Undef:
        frame = getframe().f_back

    if frame is None:
        raise NameError("name '%s' not defined in calling frames" % name)

    f_back = frame.f_back

    if isinstance(name, (tuple, list)):
        return [ find_ctx_var(n, count, f_back, caching) for n in name ]

    try:
        result = frame.f_locals[name]

        if count <= 1: return result

        return find_ctx_var(name, count-1, f_back, caching)

    except KeyError:
        r = find_ctx_var(name, count, f_back, caching)

        if caching:
            # do this for caching. Maybe this is semantically incorrect 
            frame.f_locals[name] = r
        return r

def ctx(name, value=Undef, caching=None, back=1):
    '''set or get a context variable

    :param name:
        name of variable
    :type name:
        String

    :param value:
        Any value the variable shall be set to. Default is Undef, a local
        class, which indicates that rather a value shall be retrieved than
        stored.

        You may also set the value to Here, this will still retrieve the
        value but will create a copy of the variable in current frame.

    :param caching:
        Set caching on, if there shall be created a copy in each frame
        between current one and the one holding the variable. Similar to
        *value* = *Here*.

    :param back:
        Number of frames initially go back. This usually needs not to be
        set and is rather for internal calls, see :func:`set`
    '''

    if caching is None:
        caching = ctx.CACHING

    frame = getframe()
    while back > 0:
        frame = frame.f_back
        back -= 1

    if value is Undef:
        return find_ctx_var(name, frame = frame, caching = caching)

    if value is Here:
        if isinstance(name, (tuple, list)):
            r = []
            for n in name:
                x = find_ctx_var(n, frame = frame, caching = caching)
                frame.f_locals[n] = x
                    
                r.append(x)
        else:
            r = find_ctx_var(name, frame = frame, caching = caching)
            frame.f_locals[name] = r
        return r
    else:
        frame.f_locals[name] = value

ctx.CACHING = None

def ctxget(name, default=None):
    '''get a context variable. If not exists return default value'''
    try:
        return ctx(name)
    except NameError:
        return default

get = ctxget

def set(name, value):
    '''set context variable name = value'''
    return ctx(name, value, back=2)

def _test():
    import doctest
    doctest.testmod()

def ctxstr(s, default=""):
    '''expand each ${...} s to its ctxvar. if default is None, replace by match'''
    import re
    return re.sub(r"\$\{([^\}\s]+)\}", 
              lambda m: ctxget(m.group(1), default is None and m.group(0) or default), s)

if __name__ == "__main__":
    _test()
    
